@@ -23,12 +23,14 @@ import gc
 import io
 import logging
 import os
+import platform
 import pprint
 import shutil
+import stat
 import sys
 import time
 
-from contextlib import closing
+from contextlib import closing, contextmanager
 from functools import wraps
 from tempfile import mkstemp, mkdtemp
 from threading import Thread
@@ -66,6 +68,14 @@ logging.basicConfig(
 )
 
 
+requires_unprivileged_user = pytest.mark.skipif(
+    os.geteuid() == 0, reason="This test can not run as root")
+
+
+def on_s390x():
+    return platform.machine() == "s390x"
+
+
 def skip_in_valgrind(f):
     @wraps(f)
     def wrapper(*args, **kwargs):
@@ -79,29 +89,6 @@ def skip_in_valgrind(f):
 
 class IOProcessTests(TestCase):
 
-    def testMaxRequests(self):
-        proc = IOProcess(timeout=10, max_threads=1, max_queued_requests=1)
-        with closing(proc):
-            t1 = Thread(target=proc.echo, args=("hello", 2))
-            t2 = Thread(target=proc.echo, args=("hello", 2))
-            t1.start()
-            t2.start()
-            # Make sure the echo calls are sent prior to the ping otherwise one
-            # of them would fail and ping() would pass
-            time.sleep(0.5)
-
-            try:
-                proc.ping()
-            except OSError as e:
-                self.assertEquals(e.errno, errno.EAGAIN)
-            except Exception:
-                self.fail("Expected OSError got %s", e)
-            else:
-                self.fail("Expected exception")
-            finally:
-                t1.join()
-                t2.join()
-
     def testMaxRequestsAfterFillingThreadPool(self):
         proc = IOProcess(timeout=10, max_threads=3, max_queued_requests=0)
         with closing(proc):
@@ -256,6 +243,7 @@ class IOProcessTests(TestCase):
             self.assertTrue(RSSDiff < acceptableRSSIncreasKB,
                             "Detected a leak sized %d KB" % RSSDiff)
 
+    @pytest.mark.xfail(on_s390x(), reason="Unknown")
     def testStatvfs(self):
         data = b'''Peter Puppy: Once again, evil is as rotting meat before
                                 the maggots of justice!
@@ -448,16 +436,6 @@ class IOProcessTests(TestCase):
             finally:
                 os.unlink(path)
 
-    def testAccess(self):
-        proc = IOProcess(timeout=10, max_threads=5)
-        with closing(proc):
-            fd, path = mkstemp()
-            try:
-                os.close(fd)
-                self.assertTrue(proc.access(path, os.W_OK))
-            finally:
-                os.unlink(path)
-
     def testChmodFail(self):
         proc = IOProcess(timeout=10, max_threads=5)
         with closing(proc):
@@ -584,14 +562,24 @@ class IOProcessTests(TestCase):
                 self.assertNotIn(my_file.fileno(), child_fds)
 
 
-def test_fsyncpath_file(tmpdir):
-    proc = IOProcess(timeout=10, max_threads=1)
+def test_max_requests():
+    proc = IOProcess(timeout=10, max_threads=1, max_queued_requests=1)
     with closing(proc):
-        path = tmpdir.join("file")
-        path.write("data")
-        # No easy way to test that we actually fsync this path. Lets just
-        # call it to make sure it does not fail.
-        assert proc.fsyncPath(str(path)) is None
+        t1 = Thread(target=proc.echo, args=("hello", 2))
+        t2 = Thread(target=proc.echo, args=("hello", 2))
+        t1.start()
+        t2.start()
+        # Make sure the echo calls are sent prior to the ping otherwise one
+        # of them would fail and ping() would pass.
+        time.sleep(0.5)
+
+        try:
+            with pytest.raises(OSError) as e:
+                proc.ping()
+            assert e.value.errno == errno.EAGAIN
+        finally:
+            t1.join()
+            t2.join()
 
 
 def test_fsyncpath_directory(tmpdir):
@@ -599,7 +587,7 @@ def test_fsyncpath_directory(tmpdir):
     with closing(proc):
         # No easy way to test that we actually fsync this path. Lets just
         # call it to make sure it does not fail.
-        assert proc.fsyncPath(str(tmpdir)) is None
+        proc.fsyncPath(str(tmpdir))
 
 
 def test_fsyncpath_missing(tmpdir):
@@ -709,9 +697,18 @@ def test_writefile(tmpdir, size):
         assert written == data
 
 
-# TODO: Assumes storage with 512 bytes sector size. Need to test with 4k
-# storage.
-@pytest.mark.parametrize("size", [0, 512, 1024**2])
+# TODO: Use userstorage instead of assuming CI storage sector size.
+@pytest.mark.parametrize("size", [
+    0,
+    pytest.param(
+        512,
+        marks=pytest.mark.xfail(
+            on_s390x(),
+            reason="Test assumes 512 block size but on out s390x slave "
+                   "storage uses 4k sector size")),
+    4096,
+    1024**2
+])
 def test_writefile_direct_aligned(tmpdir, size):
     data = b'x' * size
     proc = IOProcess(timeout=10, max_threads=5)
@@ -760,14 +757,168 @@ def test_readfile_direct(tmpdir, size):
         assert read == data
 
 
-class TestWeakerf(TestCase):
+@pytest.mark.parametrize("direct", [
+    pytest.param(True, id="direct"),
+    pytest.param(False, id="buffered"),
+])
+def test_readfile_missing(tmpdir, direct):
+    path = str(tmpdir.join("file"))
+    proc = IOProcess(timeout=10, max_threads=5)
+    with closing(proc):
+        with pytest.raises(OSError) as e:
+            read = proc.readfile(path, direct=direct)
+        assert e.value.errno == errno.ENOENT
+
+
+ACCESS_PARAMS = [
+    (0o755, os.R_OK, True),
+    (0o300, os.R_OK, False),
+    (0o744, os.W_OK, True),
+    (0o444, os.W_OK, False),
+    (0o755, os.X_OK, True),
+    (0o400, os.X_OK, False),
+    (0o300, os.W_OK | os.X_OK, True),
+    (0o300, os.R_OK | os.W_OK | os.X_OK, False),
+]
+
+
+@pytest.mark.parametrize("mode, permission, expected_result", ACCESS_PARAMS)
+@requires_unprivileged_user
+def test_access_file(tmpdir, mode, permission, expected_result):
+    proc = IOProcess(timeout=10, max_threads=5)
+    f = tmpdir.join("file")
+    f.write("")
+    path = str(f)
+
+    with closing(proc):
+        with chmod(path, mode):
+            assert proc.access(path, permission) == os.access(path, permission)
+            assert proc.access(path, permission) == expected_result
+
+
+@pytest.mark.parametrize("mode, permission, expected_result", ACCESS_PARAMS)
+@requires_unprivileged_user
+def test_access_directory(tmpdir, mode, permission, expected_result):
+    proc = IOProcess(timeout=10, max_threads=5)
+    d = tmpdir.mkdir("subdir")
+    path = str(d)
+
+    with closing(proc):
+        with chmod(path, mode):
+            assert proc.access(path, permission) == os.access(path, permission)
+            assert proc.access(path, permission) == expected_result
+
+
+# TODO: Use userstorage instead of assuming block size.
+@pytest.mark.xfail(
+    on_s390x(),
+    reason="Assuming 512, on s390x we have 4k drives")
+def test_probe_block_size_512(tmpdir):
+    proc = IOProcess(timeout=10, max_threads=5)
+    with closing(proc):
+        assert proc.probe_block_size(str(tmpdir)) == 512
+
+
+# TODO: Use userstorage instead of assuming block size.
+@pytest.mark.skipif(not on_s390x(), reason="Assuming block size 512")
+@pytest.mark.xfail(reason="Test assumes 4k block size on s390x")
+def test_probe_block_size_4096(tmpdir):
+    proc = IOProcess(timeout=10, max_threads=5)
+    with closing(proc):
+        assert proc.probe_block_size(str(tmpdir)) == 4096
+
+
+def test_probe_block_size_unsupported():
+    proc = IOProcess(timeout=10, max_threads=5)
+    with closing(proc):
+        # /dev/shm uses tempfs, does not support direct I/O.
+        with pytest.raises(OSError) as e:
+            proc.probe_block_size("/dev/shm")
+        assert e.value.errno == errno.EINVAL
+
+
+def test_probe_block_size_cleanup(tmpdir):
+    proc = IOProcess(timeout=10, max_threads=5)
+    with closing(proc):
+        proc.probe_block_size(str(tmpdir))
+    # Should not leave the probe temporary file around.
+    assert tmpdir.listdir() == []
+
+
+@requires_unprivileged_user
+def test_probe_block_size_not_writable(tmpdir):
+    no_write = str(tmpdir.mkdir("no-write-for-you"))
+    # Remove write bit, so the probe file cannot be created.
+    with chmod(no_write, 0o500):
+        proc = IOProcess(timeout=10, max_threads=5)
+        with closing(proc):
+            with pytest.raises(OSError) as e:
+                proc.probe_block_size(no_write)
+            assert e.value.errno == errno.EACCES
+
+
+def test_probe_block_size_concurrent(tmpdir):
+    proc = IOProcess(timeout=10, max_threads=20)
+    with closing(proc):
+        probe_dir = str(tmpdir)
+        threads = []
+        results = []
+
+        def worker():
+            for i in range(20):
+                res = proc.probe_block_size(probe_dir)
+                results.append(res)
+
+        try:
+            for i in range(20):
+                t = Thread(target=worker)
+                t.deamon = True
+                t.start()
+                threads.append(t)
+        finally:
+            for t in threads:
+                t.join()
+
+    # We should have identical prob results.
+    assert len(results) == 400
+    assert len(frozenset(results)) == 1
+
+    # And no leftover probe files.
+    assert tmpdir.listdir() == []
+
+
+@contextmanager
+def chmod(path, mode):
+    """Changes path permissions.
+
+    Changes the path permissions to the requested ones and
+    reverts to the original permissions at the end.
+
+    Arguments:
+        path (str): file/directory path
+        mode (int): new mode
+    """
+
+    orig_mode = stat.S_IMODE(os.stat(path).st_mode)
+    os.chmod(path, mode)
+
+    try:
+        yield
+    finally:
+        try:
+            os.chmod(path, orig_mode)
+        except Exception as e:
+            logging.error("Failed to restore %r mode: %s", path, e)
+
+
+class TestWeakref(TestCase):
 
     @pytest.mark.xfail(
         sys.version_info[:2] == (3, 7), reason="fails in python 3.7")
     def test_close_when_unrefed(self):
-        """Make sure there is nothing keepin IOProcess strongly referenced.
+        """Make sure there is nothing keeping IOProcess strongly referenced.
 
-        Since there is a comminucation background thread doing all the hard
+        Since there is a communication background thread doing all the hard
         work we need to make sure it doesn't prevent IOProcess from being
         garbage collected.
         """
